"""
    verktyg.datastructures
    ~~~~~~~~~~~~~~~~~~~~~~

    This module provides mixins and classes with an immutable interface.

    :copyright:
        (c) 2015 Ben Mather, based on Werkzeug, see AUTHORS for more details.
    :license:
        BSD, see LICENSE for more details.
"""
import re
from copy import deepcopy
from itertools import repeat

from verktyg import exceptions


_missing = object()
_locale_delim_re = re.compile(r'[_-]')


def is_immutable(self):
    raise TypeError('%r objects are immutable' % self.__class__.__name__)


def iter_multi_items(mapping):
    """Iterates over the items of a mapping yielding keys and values
    without dropping any from more complex structures.
    """
    if isinstance(mapping, MultiDict):
        for item in mapping.items(multi=True):
            yield item
    elif isinstance(mapping, dict):
        for key, value in mapping.items():
            if isinstance(value, (tuple, list)):
                for value in value:
                    yield key, value
            else:
                yield key, value
    else:
        for item in mapping:
            yield item


class ImmutableListMixin(object):
    """Makes a :class:`list` immutable.

    :private:
    """

    _hash_cache = None

    def __hash__(self):
        if self._hash_cache is not None:
            return self._hash_cache
        rv = self._hash_cache = hash(tuple(self))
        return rv

    def __reduce_ex__(self, protocol):
        return type(self), (list(self),)

    def __delitem__(self, key):
        is_immutable(self)

    def __delslice__(self, i, j):
        is_immutable(self)

    def __iadd__(self, other):
        is_immutable(self)
    __imul__ = __iadd__

    def __setitem__(self, key, value):
        is_immutable(self)

    def __setslice__(self, i, j, value):
        is_immutable(self)

    def append(self, item):
        is_immutable(self)
    remove = append

    def extend(self, iterable):
        is_immutable(self)

    def insert(self, pos, value):
        is_immutable(self)

    def pop(self, index=-1):
        is_immutable(self)

    def reverse(self):
        is_immutable(self)

    def sort(self, cmp=None, key=None, reverse=None):
        is_immutable(self)


class ImmutableList(ImmutableListMixin, list):
    """An immutable :class:`list`.

    :private:
    """

    def __repr__(self):
        return '%s(%s)' % (
            self.__class__.__name__,
            list.__repr__(self),
        )


class ImmutableDictMixin(object):
    """Makes a :class:`dict` immutable.

    :private:
    """
    _hash_cache = None

    @classmethod
    def fromkeys(cls, keys, value=None):
        instance = super(cls, cls).__new__(cls)
        instance.__init__(zip(keys, repeat(value)))
        return instance

    def __reduce_ex__(self, protocol):
        return type(self), (dict(self),)

    def _iter_hashitems(self):
        return self.items()

    def __hash__(self):
        if self._hash_cache is not None:
            return self._hash_cache
        rv = self._hash_cache = hash(frozenset(self._iter_hashitems()))
        return rv

    def setdefault(self, key, default=None):
        is_immutable(self)

    def update(self, *args, **kwargs):
        is_immutable(self)

    def pop(self, key, default=None):
        is_immutable(self)

    def popitem(self):
        is_immutable(self)

    def __setitem__(self, key, value):
        is_immutable(self)

    def __delitem__(self, key):
        is_immutable(self)

    def clear(self):
        is_immutable(self)


class ImmutableMultiDictMixin(ImmutableDictMixin):
    """Makes a :class:`MultiDict` immutable.

    :private:
    """

    def __reduce_ex__(self, protocol):
        return type(self), (list(self.items(multi=True)),)

    def _iter_hashitems(self):
        return self.items(multi=True)

    def add(self, key, value):
        is_immutable(self)

    def popitemlist(self):
        is_immutable(self)

    def poplist(self, key):
        is_immutable(self)

    def setlist(self, key, new_list):
        is_immutable(self)

    def setlistdefault(self, key, default_list=None):
        is_immutable(self)


class UpdateDictMixin(object):
    """Makes dicts call `self.on_update` on modifications.

    :private:
    """

    on_update = None

    def calls_update(name):
        def oncall(self, *args, **kw):
            rv = getattr(super(UpdateDictMixin, self), name)(*args, **kw)
            if self.on_update is not None:
                self.on_update(self)
            return rv
        oncall.__name__ = name
        return oncall

    def setdefault(self, key, default=None):
        modified = key not in self
        rv = super(UpdateDictMixin, self).setdefault(key, default)
        if modified and self.on_update is not None:
            self.on_update(self)
        return rv

    def pop(self, key, default=_missing):
        modified = key in self
        if default is _missing:
            rv = super(UpdateDictMixin, self).pop(key)
        else:
            rv = super(UpdateDictMixin, self).pop(key, default)
        if modified and self.on_update is not None:
            self.on_update(self)
        return rv

    __setitem__ = calls_update('__setitem__')
    __delitem__ = calls_update('__delitem__')
    clear = calls_update('clear')
    popitem = calls_update('popitem')
    update = calls_update('update')
    del calls_update


class TypeConversionDict(dict):
    """Works like a regular dict but the :meth:`get` method can perform
    type conversions.  :class:`MultiDict` and :class:`CombinedMultiDict`
    are subclasses of this class and provide the same feature.
    """

    def get(self, key, default=None, type=None):
        """Return the default value if the requested data doesn't exist.
        If `type` is provided and is a callable it should convert the value,
        return it or raise a :exc:`ValueError` if that is not possible.  In
        this case the function will return the default as if the value was not
        found:

        >>> d = TypeConversionDict(foo='42', bar='blub')
        >>> d.get('foo', type=int)
        42
        >>> d.get('bar', -1, type=int)
        -1

        :param key:
            The key to be looked up.
        :param default:
            The default value to be returned if the key can't be looked up.  If
            not further specified `None` is returned.
        :param type:
            A callable that is used to cast the value in the
            :class:`MultiDict`.  If a :exc:`ValueError` is raised by this
            callable the default value is returned.
        """
        try:
            rv = self[key]
            if type is not None:
                rv = type(rv)
        except (KeyError, ValueError):
            rv = default
        return rv


class ImmutableTypeConversionDict(ImmutableDictMixin, TypeConversionDict):
    """Works like a :class:`TypeConversionDict` but does not support
    modifications.
    """

    def copy(self):
        """Return a shallow mutable copy of this object.  Keep in mind that
        the standard library's :func:`copy` function is a no-op for this class
        like for any other python immutable type (eg: :class:`tuple`).
        """
        return TypeConversionDict(self)

    def __copy__(self):
        return self


class MultiDict(TypeConversionDict):
    """A :class:`MultiDict` is a dictionary subclass customized to deal with
    multiple values for the same key which is for example used by the parsing
    functions in the wrappers.  This is necessary because some HTML form
    elements pass multiple values for the same key.

    :class:`MultiDict` implements all standard dictionary methods.
    Internally, it saves all values for a key as a list, but the standard dict
    access methods will only return the first value for a key. If you want to
    gain access to the other values, too, you have to use the `list` methods as
    explained below.

    Basic Usage:

    >>> d = MultiDict([('a', 'b'), ('a', 'c')])
    >>> d
    MultiDict([('a', 'b'), ('a', 'c')])
    >>> d['a']
    'b'
    >>> d.getlist('a')
    ['b', 'c']
    >>> 'a' in d
    True

    It behaves like a normal dict thus all dict functions will only return the
    first value when multiple values for one key are found.

    From Werkzeug 0.3 onwards, the `KeyError` raised by this class is also a
    subclass of the :exc:`~exceptions.BadRequest` HTTP exception and will
    render a page for a ``400 BAD REQUEST`` if caught in a catch-all for HTTP
    exceptions.

    A :class:`MultiDict` can be constructed from an iterable of
    ``(key, value)`` tuples, a dict, a :class:`MultiDict` or from Werkzeug 0.2
    onwards some keyword parameters.

    :param mapping:
        the initial value for the :class:`MultiDict`.  Either a regular dict,
        an iterable of ``(key, value)`` tuples or `None`.
    """

    def __init__(self, mapping=None):
        if isinstance(mapping, MultiDict):
            dict.__init__(self, ((k, l[:]) for k, l in mapping.lists()))
        elif isinstance(mapping, dict):
            tmp = {}
            for key, value in mapping.items():
                if isinstance(value, (tuple, list)):
                    value = list(value)
                else:
                    value = [value]
                tmp[key] = value
            dict.__init__(self, tmp)
        else:
            tmp = {}
            for key, value in mapping or ():
                tmp.setdefault(key, []).append(value)
            dict.__init__(self, tmp)

    def __getstate__(self):
        return dict(self.lists())

    def __setstate__(self, value):
        dict.clear(self)
        dict.update(self, value)

    def __getitem__(self, key):
        """Return the first data value for this key;
        raises KeyError if not found.

        :param key:
            The key to be looked up.
        :raise KeyError:
            If the key does not exist.
        """
        if key in self:
            return dict.__getitem__(self, key)[0]
        raise exceptions.BadRequestKeyError(key)

    def __setitem__(self, key, value):
        """Like :meth:`add` but removes an existing key first.

        :param key:
            the key for the value.
        :param value:
            the value to set.
        """
        dict.__setitem__(self, key, [value])

    def add(self, key, value):
        """Adds a new value for the key.

        :param key:
            The key for the value.
        :param value:
            The value to add.
        """
        dict.setdefault(self, key, []).append(value)

    def getlist(self, key, type=None):
        """Return the list of items for a given key. If that key is not in the
        `MultiDict`, the return value will be an empty list.  Just as `get`
        `getlist` accepts a `type` parameter.  All items will be converted
        with the callable defined there.

        :param key:
            The key to be looked up.
        :param type:
            A callable that is used to cast the value in the
            :class:`MultiDict`.  If a :exc:`ValueError` is raised by this
            callable the value will be removed from the list.
        :return: a :class:`list` of all the values for the key.
        """
        try:
            rv = dict.__getitem__(self, key)
        except KeyError:
            return []
        if type is None:
            return list(rv)
        result = []
        for item in rv:
            try:
                result.append(type(item))
            except ValueError:
                pass
        return result

    def setlist(self, key, new_list):
        """Remove the old values for a key and add new ones.  Note that the
        list you pass the values in will be shallow-copied before it is
        inserted in the dictionary.

        >>> d = MultiDict()
        >>> d.setlist('foo', ['1', '2'])
        >>> d['foo']
        '1'
        >>> d.getlist('foo')
        ['1', '2']

        :param key:
            The key for which the values are set.
        :param new_list:
            An iterable with the new values for the key.  Old values are
            removed first.
        """
        dict.__setitem__(self, key, list(new_list))

    def setdefault(self, key, default=None):
        """Returns the value for the key if it is in the dict, otherwise it
        returns `default` and sets that value for `key`.

        :param key:
            The key to be looked up.
        :param default:
            The default value to be returned if the key is not in the dict.  If
            not further specified it's `None`.
        """
        if key not in self:
            self[key] = default
        else:
            default = self[key]
        return default

    def setlistdefault(self, key, default_list=None):
        """Like `setdefault` but sets multiple values.  The list returned
        is not a copy, but the list that is actually used internally.  This
        means that you can put new values into the dict by appending items
        to the list:

        >>> d = MultiDict({"foo": 1})
        >>> d.setlistdefault("foo").extend([2, 3])
        >>> d.getlist("foo")
        [1, 2, 3]

        :param key:
            The key to be looked up.
        :param default:
            An iterable of default values.  It is either copied (in case it was
            a list) or converted into a list before returned.
        :return:
            a :class:`list`
        """
        if key not in self:
            default_list = list(default_list or ())
            dict.__setitem__(self, key, default_list)
        else:
            default_list = dict.__getitem__(self, key)
        return default_list

    def items(self, multi=False):
        """Return an iterator of ``(key, value)`` pairs.

        :param multi:
            If set to `True` the iterator returned will have a pair for each
            value of each key.  Otherwise it will only contain pairs for the
            first value of each key.
        """

        for key, values in dict.items(self):
            if multi:
                for value in values:
                    yield key, value
            else:
                yield key, values[0]

    def lists(self):
        """Return a list of ``(key, values)`` pairs, where values is the list
        of all values associated with the key."""

        for key, values in dict.items(self):
            yield key, list(values)

    def keys(self):
        return dict.keys(self)

    def __iter__(self):
        return iter(self.keys())

    def values(self):
        """Returns an iterator of the first value on every key's value list."""
        for values in dict.values(self):
            yield values[0]

    def listvalues(self):
        """Return an iterator of all values associated with a key.  Zipping
        :meth:`keys` and this is the same as calling :meth:`lists`:

        >>> d = MultiDict({"foo": [1, 2, 3]})
        >>> zip(d.keys(), d.listvalues()) == d.lists()
        True
        """

        return dict.values(self)

    def copy(self):
        """Return a shallow copy of this object."""
        return self.__class__(self)

    def deepcopy(self, memo=None):
        """Return a deep copy of this object."""
        return self.__class__(deepcopy(self.to_dict(flat=False), memo))

    def to_dict(self, flat=True):
        """Return the contents as regular dict.  If `flat` is `True` the
        returned dict will only have the first item present, if `flat` is
        `False` all values will be returned as lists.

        :param flat:
            If set to `False` the dict returned will have lists with all the
            values in it.  Otherwise it will only contain the first value for
            each key.
        :return:
            A :class:`dict`
        """
        if flat:
            return dict(self.items())
        return dict(self.lists())

    def update(self, other_dict):
        """update() extends rather than replaces existing key lists."""
        for key, value in iter_multi_items(other_dict):
            MultiDict.add(self, key, value)

    def pop(self, key, default=_missing):
        """Pop the first item for a list on the dict.  Afterwards the
        key is removed from the dict, so additional values are discarded:

        >>> d = MultiDict({"foo": [1, 2, 3]})
        >>> d.pop("foo")
        1
        >>> "foo" in d
        False

        :param key:
            The key to pop.
        :param default:
            If provided the value to return if the key was not in the
            dictionary.
        """
        try:
            return dict.pop(self, key)[0]
        except KeyError as e:
            if default is not _missing:
                return default
            raise exceptions.BadRequestKeyError(str(e))

    def popitem(self):
        """Pop an item from the dict."""
        try:
            item = dict.popitem(self)
            return (item[0], item[1][0])
        except KeyError as e:
            raise exceptions.BadRequestKeyError(str(e))

    def poplist(self, key):
        """Pop the list for a key from the dict.  If the key is not in the dict
        an empty list is returned.
        """
        return dict.pop(self, key, [])

    def popitemlist(self):
        """Pop a ``(key, list)`` tuple from the dict."""
        try:
            return dict.popitem(self)
        except KeyError as e:
            raise exceptions.BadRequestKeyError(str(e))

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.deepcopy(memo=memo)

    def __repr__(self):
        return '%s(%r)' % (
            self.__class__.__name__, list(self.items(multi=True))
        )


class _omd_bucket(object):
    """Wraps values in the :class:`OrderedMultiDict`.  This makes it
    possible to keep an order over multiple different keys.  It requires
    a lot of extra memory and slows down access a lot, but makes it
    possible to access elements in O(1) and iterate in O(n).
    """
    __slots__ = ('prev', 'key', 'value', 'next')

    def __init__(self, omd, key, value):
        self.prev = omd._last_bucket
        self.key = key
        self.value = value
        self.next = None

        if omd._first_bucket is None:
            omd._first_bucket = self
        if omd._last_bucket is not None:
            omd._last_bucket.next = self
        omd._last_bucket = self

    def unlink(self, omd):
        if self.prev:
            self.prev.next = self.next
        if self.next:
            self.next.prev = self.prev
        if omd._first_bucket is self:
            omd._first_bucket = self.next
        if omd._last_bucket is self:
            omd._last_bucket = self.prev


class OrderedMultiDict(MultiDict):
    """Works like a regular :class:`MultiDict` but preserves the
    order of the fields.  To convert the ordered multi dict into a
    list you can use the :meth:`items` method and pass it ``multi=True``.

    In general an :class:`OrderedMultiDict` is an order of magnitude
    slower than a :class:`MultiDict`.

    .. admonition:: note

       Due to a limitation in Python you cannot convert an ordered
       multi dict into a regular dict by using ``dict(multidict)``.
       Instead you have to use the :meth:`to_dict` method, otherwise
       the internal bucket objects are exposed.
    """

    def __init__(self, mapping=None):
        dict.__init__(self)
        self._first_bucket = self._last_bucket = None
        if mapping is not None:
            OrderedMultiDict.update(self, mapping)

    def __eq__(self, other):
        if not isinstance(other, MultiDict):
            return NotImplemented
        if isinstance(other, OrderedMultiDict):
            iter1 = self.items(multi=True)
            iter2 = other.items(multi=True)
            try:
                for k1, v1 in iter1:
                    k2, v2 = next(iter2)
                    if k1 != k2 or v1 != v2:
                        return False
            except StopIteration:
                return False
            try:
                next(iter2)
            except StopIteration:
                return True
            return False
        if len(self) != len(other):
            return False
        for key, values in self.lists():
            if other.getlist(key) != values:
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __reduce_ex__(self, protocol):
        return type(self), (list(self.items(multi=True)),)

    def __getstate__(self):
        return list(self.items(multi=True))

    def __setstate__(self, values):
        dict.clear(self)
        for key, value in values:
            self.add(key, value)

    def __getitem__(self, key):
        if key in self:
            return dict.__getitem__(self, key)[0].value
        raise exceptions.BadRequestKeyError(key)

    def __setitem__(self, key, value):
        self.poplist(key)
        self.add(key, value)

    def __delitem__(self, key):
        self.pop(key)

    def keys(self):
        return (key for key, value in self.items())

    __iter__ = keys

    def values(self):
        return (value for key, value in self.items())

    def items(self, multi=False):
        ptr = self._first_bucket
        if multi:
            while ptr is not None:
                yield ptr.key, ptr.value
                ptr = ptr.next
        else:
            returned_keys = set()
            while ptr is not None:
                if ptr.key not in returned_keys:
                    returned_keys.add(ptr.key)
                    yield ptr.key, ptr.value
                ptr = ptr.next

    def lists(self):
        returned_keys = set()
        ptr = self._first_bucket
        while ptr is not None:
            if ptr.key not in returned_keys:
                yield ptr.key, self.getlist(ptr.key)
                returned_keys.add(ptr.key)
            ptr = ptr.next

    def listvalues(self):
        for key, values in self.lists():
            yield values

    def add(self, key, value):
        dict.setdefault(self, key, []).append(_omd_bucket(self, key, value))

    def getlist(self, key, type=None):
        try:
            rv = dict.__getitem__(self, key)
        except KeyError:
            return []
        if type is None:
            return [x.value for x in rv]
        result = []
        for item in rv:
            try:
                result.append(type(item.value))
            except ValueError:
                pass
        return result

    def setlist(self, key, new_list):
        self.poplist(key)
        for value in new_list:
            self.add(key, value)

    def setlistdefault(self, key, default_list=None):
        raise TypeError(
            'setlistdefault is unsupported for ordered multi dicts'
        )

    def update(self, mapping):
        for key, value in iter_multi_items(mapping):
            OrderedMultiDict.add(self, key, value)

    def poplist(self, key):
        buckets = dict.pop(self, key, ())
        for bucket in buckets:
            bucket.unlink(self)
        return [x.value for x in buckets]

    def pop(self, key, default=_missing):
        try:
            buckets = dict.pop(self, key)
        except KeyError as e:
            if default is not _missing:
                return default
            raise exceptions.BadRequestKeyError(str(e))
        for bucket in buckets:
            bucket.unlink(self)
        return buckets[0].value

    def popitem(self):
        try:
            key, buckets = dict.popitem(self)
        except KeyError as e:
            raise exceptions.BadRequestKeyError(str(e))
        for bucket in buckets:
            bucket.unlink(self)
        return key, buckets[0].value

    def popitemlist(self):
        try:
            key, buckets = dict.popitem(self)
        except KeyError as e:
            raise exceptions.BadRequestKeyError(str(e))
        for bucket in buckets:
            bucket.unlink(self)
        return key, [x.value for x in buckets]


class CombinedMultiDict(ImmutableMultiDictMixin, MultiDict):
    """A read only :class:`MultiDict` that you can pass multiple
    :class:`MultiDict` instances as sequence and it will combine the return
    values of all wrapped dicts:

    >>> from verktyg.datastructures import CombinedMultiDict, MultiDict
    >>> post = MultiDict([('foo', 'bar')])
    >>> get = MultiDict([('blub', 'blah')])
    >>> combined = CombinedMultiDict([get, post])
    >>> combined['foo']
    'bar'
    >>> combined['blub']
    'blah'

    This works for all read operations and will raise a `TypeError` for
    methods that usually change data which isn't possible.

    From Werkzeug 0.3 onwards, the `KeyError` raised by this class is also a
    subclass of the :exc:`~exceptions.BadRequest` HTTP exception and will
    render a page for a ``400 BAD REQUEST`` if caught in a catch-all for HTTP
    exceptions.
    """

    def __reduce_ex__(self, protocol):
        return type(self), (self.dicts,)

    def __init__(self, dicts=None):
        self.dicts = dicts or []

    @classmethod
    def fromkeys(cls):
        raise TypeError((
            'cannot create {cls!r} instances by fromkeys'
        ).format(cls=cls.__name__))

    def __getitem__(self, key):
        for d in self.dicts:
            if key in d:
                return d[key]
        raise exceptions.BadRequestKeyError(key)

    def get(self, key, default=None, type=None):
        for d in self.dicts:
            if key in d:
                if type is not None:
                    try:
                        return type(d[key])
                    except ValueError:
                        continue
                return d[key]
        return default

    def getlist(self, key, type=None):
        rv = []
        for d in self.dicts:
            rv.extend(d.getlist(key, type))
        return rv

    def _keys_impl(self):
        """This function exists so __len__ can be implemented more efficiently,
        saving one list creation from an iterator.

        Using this for Python 2's ``dict.keys`` behavior would be useless since
        `dict.keys` in Python 2 returns a list, while we have a set here.
        """
        rv = set()
        for d in self.dicts:
            rv.update(d.keys())
        return rv

    def keys(self):
        return iter(self._keys_impl())

    __iter__ = keys

    def items(self, multi=False):
        found = set()
        for d in self.dicts:
            for key, value in d.items(multi):
                if multi:
                    yield key, value
                elif key not in found:
                    found.add(key)
                    yield key, value

    def values(self):
        for key, value in self.items():
            yield value

    def lists(self):
        rv = {}
        for d in self.dicts:
            for key, values in d.lists():
                rv.setdefault(key, []).extend(values)
        return rv.items()

    def listvalues(self):
        return (x[1] for x in self.lists())

    def copy(self):
        """Return a shallow copy of this object."""
        return self.__class__(self.dicts[:])

    def to_dict(self, flat=True):
        """Return the contents as regular dict.  If `flat` is `True` the
        returned dict will only have the first item present, if `flat` is
        `False` all values will be returned as lists.

        :param flat:
            If set to `False` the dict returned will have lists with all the
            values in it.  Otherwise it will only contain the first item for
            each key.
        :return:
            A :class:`dict`
        """
        rv = {}
        for d in reversed(self.dicts):
            rv.update(d.to_dict(flat))
        return rv

    def __len__(self):
        return len(self._keys_impl())

    def __contains__(self, key):
        for d in self.dicts:
            if key in d:
                return True
        return False

    has_key = __contains__

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.dicts)


class ImmutableDict(ImmutableDictMixin, dict):
    """An immutable :class:`dict`.
    """

    def __repr__(self):
        return '%s(%s)' % (
            self.__class__.__name__,
            dict.__repr__(self),
        )

    def copy(self):
        """Return a shallow mutable copy of this object.  Keep in mind that
        the standard library's :func:`copy` function is a no-op for this class
        like for any other python immutable type (eg: :class:`tuple`).
        """
        return dict(self)

    def __copy__(self):
        return self


class ImmutableMultiDict(ImmutableMultiDictMixin, MultiDict):
    """An immutable :class:`MultiDict`.
    """

    def copy(self):
        """Return a shallow mutable copy of this object.  Keep in mind that
        the standard library's :func:`copy` function is a no-op for this class
        like for any other python immutable type (eg: :class:`tuple`).
        """
        return MultiDict(self)

    def __copy__(self):
        return self


class ImmutableOrderedMultiDict(ImmutableMultiDictMixin, OrderedMultiDict):
    """An immutable :class:`OrderedMultiDict`.
    """

    def _iter_hashitems(self):
        return enumerate(self.items(multi=True))

    def copy(self):
        """Return a shallow mutable copy of this object.  Keep in mind that
        the standard library's :func:`copy` function is a no-op for this class
        like for any other python immutable type (eg: :class:`tuple`).
        """
        return OrderedMultiDict(self)

    def __copy__(self):
        return self


class CallbackDict(UpdateDictMixin, dict):
    """A dict that calls a function passed every time something is changed.
    The function is passed the dict instance.
    """

    def __init__(self, initial=None, on_update=None):
        dict.__init__(self, initial or ())
        self.on_update = on_update

    def __repr__(self):
        return '<%s %s>' % (
            self.__class__.__name__,
            dict.__repr__(self)
        )



"""
    verktyg.accept._base
    ~~~~~~~~~~~~~~~~~~~~

    :copyright:
        (c) 2016 Ben Mather
    :license:
        BSD, see LICENSE for more details.
"""
import re
import functools

from verktyg.datastructures import ImmutableDict
from verktyg.exceptions import NotAcceptable


_token_re_str = r'''
    (?:
        [a-zA-Z0-9]+
        (?:
            - [a-zA-Z0-9]+
        )*
    )
'''

# TODO
_value_re = re.compile(
    r'''
        ^
        {token} | \*
        $
    '''.format(token=_token_re_str), re.VERBOSE
)

# TODO
_content_type_value_re = re.compile(
    r'''
        ^
        (?: {token} | \*)
        /
        (?:
            (?:
                (?:{token} \.)?
                {token}
                (?: \+ {token})?
            ) | \*
        )
    '''.format(token=_token_re_str), re.VERBOSE
)


class Range(object):
    def __init__(self, value, q=1.0, params=None):
        self._validate_value(value)
        self.value = value

        self.q = max(min(float(q), 1.0), 0.0)

        if params is None:
            params = {}
        for param in params.items():
            self._validate_param(*param)
        self.params = ImmutableDict(params)

    def _validate_value(self, value):
        if _value_re.match(value) is None:
            raise ValueError("Invalid value: %r" % value)

    def _validate_param(self, key, value):
        return

    def __repr__(self):
        return '{name}(value={value!r}, q={q!r}, params={params!r})'.format(
            name=self.__class__.__name__,
            value=self.value,
            q=self.q,
            params=self.params,
        )

    def __str__(self):
        return self.to_header()

    def to_header(self):
        header = self.value

        if self.q != 1:
            header += ';q=%s' % self.q

        for param in self.params.items():
            # TODO escaping?
            header += ';%s=%s' % param

        return header


class Accept(object):
    range_type = None

    def __init__(self, options):
        self._options = []
        for option in options:
            if isinstance(option, str):
                option = (option,)

            self._options.append(self.range_type(*option))

    def __iter__(self):
        return iter(self._options)

    def __contains__(self, value):
        try:
            value.acceptability(self)
        except NotAcceptable:
            return False
        else:
            return True

    def __getitem__(self, value):
        try:
            return value.acceptability(self)
        except NotAcceptable as e:
            raise KeyError() from e

    def __repr__(self):
        raise NotImplementedError()

    def __str__(self):
        return self.to_header()

    def to_header(self):
        """Return an equivalent string suitable for use as an `Accept` header.
        """
        return ','.join(option.to_header() for option in self)


@functools.total_ordering
class Acceptability(object):
    def __init__(self, value, *, match_quality, q, qs=None):
        self._value = value
        self._match_quality = match_quality
        self._q = q
        self._qs = qs

    @property
    def exact_match(self):
        return bool(self._match_quality)

    @property
    def quality(self):
        if self._qs is not None:
            return self._q * self._qs
        return self._q

    def __eq__(self, other):
        if other is None:
            return False

        if self._match_quality != other._match_quality:
            return False

        if self.quality != other.quality:
            return False

        return True

    def __gt__(self, other):
        if other is None:
            return True

        if self._match_quality > other._match_quality:
            return True

        if self.quality > other.quality:
            return True

        return False


# TODO better name
class Value(object):
    """Base class for a value that the server can propose during content type
    negotiation.
    """
    match_type = None

    def __init__(self, value, *, qs=None) -> None:
        self.value = value
        self.qs = qs

    def _acceptability_for_option(self, option):
        if option.value == '*':
            exact_match = False
        elif self.value == option.value:
            exact_match = True
        else:
            raise NotAcceptable()

        return self.match_type(
            self, exact_match=exact_match,
            q=option.q, qs=self.qs
        )

    def acceptability(self, accept: Accept):
        best_match = None

        for option in accept:
            try:
                match = self._acceptability_for_option(option)
            except NotAcceptable:
                pass
            else:
                if best_match is None or match > best_match:
                    best_match = match

        if best_match is None:
            raise NotAcceptable()

        return best_match

    def __str__(self):
        return self.to_header()

    def to_header(self):
        """Returns a string suitable for use in the corresponding header
        """
        return self.value


def split_accept_string(string):
    for accept_range in string.split(','):
        accept, *str_params = accept_range.split(';')

        params = {}
        for param in str_params:
            try:
                key, value = param.split('=', 1)
            except ValueError as e:
                raise ValueError("invalid parameter: %r" % param) from e

            key, value = key.strip(), value.strip()

            params[key] = value

        q = params.pop('q', '1.0')

        yield accept.strip(), q, params



"""
Verktyg
=======

A web framework based on Werkzeug.
"""
from setuptools import setup, find_packages

extras_require = {
}

setup(
    name='verktyg',
    version='0.9.0',
    url='https://github.com/bwhmather/verktyg',
    license='BSD',
    author='Ben Mather',
    author_email='bwhmather@bwhmather.com',
    description='A web framework based on Werkzeug',
    long_description=__doc__,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    platforms='any',
    install_requires=[
    ],
    tests_require=list(set(sum(
        (extras_require[extra] for extra in {}), []
    ))),
    extras_require=extras_require,
    packages=find_packages(),
    include_package_data=True,
    test_suite='verktyg.testsuite.suite',
)



"""
    verktyg.testsuite.test_datastructures
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Tests the functionality of the provided Verktyg datastructures.

    Classes prefixed with an underscore are mixins and are not discovered by
    the test runner.

    TODO:

    -   FileMultiDict
    -   Immutable types undertested
    -   Split up dict tests

    :copyright:
        (c) 2015 Ben Mather, based on Werkzeug, see AUTHORS for more details.
    :license:
        BSD, see LICENSE for more details.

"""
import unittest


import pickle
from contextlib import contextmanager
from copy import copy, deepcopy

from verktyg import datastructures
from verktyg.exceptions import BadRequestKeyError


class NativeItermethodsTestCase(unittest.TestCase):

    def test_basic(self):
        class StupidDict(object):

            def keys(self, multi=1):
                return iter(['a', 'b', 'c'] * multi)

            def values(self, multi=1):
                return iter([1, 2, 3] * multi)

            def items(self, multi=1):
                return iter(
                    zip(self.keys(multi=multi), self.values(multi=multi))
                )

        d = StupidDict()
        expected_keys = ['a', 'b', 'c']
        expected_values = [1, 2, 3]
        expected_items = list(zip(expected_keys, expected_values))

        self.assertEqual(list(d.keys()), expected_keys)
        self.assertEqual(list(d.values()), expected_values)
        self.assertEqual(list(d.items()), expected_items)

        self.assertEqual(list(d.keys(2)), expected_keys * 2)
        self.assertEqual(list(d.values(2)), expected_values * 2)
        self.assertEqual(list(d.items(2)), expected_items * 2)


class _MutableMultiDictTestsMixin(object):
    storage_class = None

    def test_pickle(self):
        cls = self.storage_class

        def create_instance(module=None):
            if module is None:
                d = cls()
            else:
                old = cls.__module__
                cls.__module__ = module
                d = cls()
                cls.__module__ = old
            d.setlist(b'foo', [1, 2, 3, 4])
            d.setlist(b'bar', b'foo bar baz'.split())
            return d

        for protocol in range(pickle.HIGHEST_PROTOCOL + 1):
            d = create_instance()
            s = pickle.dumps(d, protocol)
            ud = pickle.loads(s)
            self.assertEqual(type(ud), type(d))
            self.assertEqual(ud, d)
            alternative = pickle.dumps(create_instance('verktyg'), protocol)
            self.assertEqual(pickle.loads(alternative), d)
            ud[b'newkey'] = b'bla'
            self.assertNotEqual(ud, d)

    def test_basic_interface(self):
        md = self.storage_class()
        self.assertIsInstance(md, dict)

        mapping = [
            ('a', 1), ('b', 2), ('a', 2), ('d', 3),
            ('a', 1), ('a', 3), ('d', 4), ('c', 3),
        ]
        md = self.storage_class(mapping)

        # simple getitem gives the first value
        self.assertEqual(md['a'], 1)
        self.assertEqual(md['c'], 3)
        self.assertRaises(KeyError, lambda: md['e'])
        self.assertEqual(md.get('a'), 1)

        # list getitem
        self.assertEqual(md.getlist('a'), [1, 2, 1, 3])
        self.assertEqual(md.getlist('d'), [3, 4])
        # do not raise if key not found
        self.assertEqual(md.getlist('x'), [])

        # simple setitem overwrites all values
        md['a'] = 42
        self.assertEqual(md.getlist('a'), [42])

        # list setitem
        md.setlist('a', [1, 2, 3])
        self.assertEqual(md['a'], 1)
        self.assertEqual(md.getlist('a'), [1, 2, 3])

        # verify that it does not change original lists
        l1 = [1, 2, 3]
        md.setlist('a', l1)
        del l1[:]
        self.assertEqual(md['a'], 1)

        # setdefault, setlistdefault
        self.assertEqual(md.setdefault('u', 23), 23)
        self.assertEqual(md.getlist('u'), [23])
        del md['u']

        md.setlist('u', [-1, -2])

        # delitem
        del md['u']
        self.assertRaises(KeyError, lambda: md['u'])
        del md['d']
        self.assertEqual(md.getlist('d'), [])

        # keys, values, items, lists
        self.assertEqual(list(sorted(md.keys())), ['a', 'b', 'c'])
        self.assertEqual(list(sorted(md.values())), [1, 2, 3])
        self.assertEqual(
            list(sorted(md.items())),
            [('a', 1), ('b', 2), ('c', 3)]
        )
        self.assertEqual(
            list(sorted(md.items(multi=True))),
            [('a', 1), ('a', 2), ('a', 3), ('b', 2), ('c', 3)]
        )
        self.assertEqual(
            list(sorted(md.lists())),
            [('a', [1, 2, 3]), ('b', [2]), ('c', [3])]
        )

        # copy method
        c = md.copy()
        self.assertEqual(c['a'], 1)
        self.assertEqual(c.getlist('a'), [1, 2, 3])

        # copy method 2
        c = copy(md)
        self.assertEqual(c['a'], 1)
        self.assertEqual(c.getlist('a'), [1, 2, 3])

        # deepcopy method
        c = md.deepcopy()
        self.assertEqual(c['a'], 1)
        self.assertEqual(c.getlist('a'), [1, 2, 3])

        # deepcopy method 2
        c = deepcopy(md)
        self.assertEqual(c['a'], 1)
        self.assertEqual(c.getlist('a'), [1, 2, 3])

        # update with a multidict
        od = self.storage_class([('a', 4), ('a', 5), ('y', 0)])
        md.update(od)
        self.assertEqual(md.getlist('a'), [1, 2, 3, 4, 5])
        self.assertEqual(md.getlist('y'), [0])

        # update with a regular dict
        md = c
        od = {'a': 4, 'y': 0}
        md.update(od)
        self.assertEqual(md.getlist('a'), [1, 2, 3, 4])
        self.assertEqual(md.getlist('y'), [0])

        # pop, poplist, popitem, popitemlist
        self.assertEqual(md.pop('y'), 0)
        self.assertNotIn('y', md)
        self.assertEqual(md.poplist('a'), [1, 2, 3, 4])
        self.assertNotIn('a', md)
        self.assertEqual(md.poplist('missing'), [])

        # remaining: b=2, c=3
        popped = md.popitem()
        self.assertIn(popped, [('b', 2), ('c', 3)])
        popped = md.popitemlist()
        self.assertIn(popped, [('b', [2]), ('c', [3])])

        # type conversion
        md = self.storage_class({'a': '4', 'b': ['2', '3']})
        self.assertEqual(md.get('a', type=int), 4)
        self.assertEqual(md.getlist('b', type=int), [2, 3])

        # repr
        md = self.storage_class([('a', 1), ('a', 2), ('b', 3)])
        self.assertIn("('a', 1)", repr(md))
        self.assertIn("('a', 2)", repr(md))
        self.assertIn("('b', 3)", repr(md))

        # add and getlist
        md.add('c', '42')
        md.add('c', '23')
        self.assertEqual(md.getlist('c'), ['42', '23'])
        md.add('c', 'blah')
        self.assertEqual(md.getlist('c', type=int), [42, 23])

        # setdefault
        md = self.storage_class()
        md.setdefault('x', []).append(42)
        md.setdefault('x', []).append(23)
        self.assertEqual(md['x'], [42, 23])

        # to dict
        md = self.storage_class()
        md['foo'] = 42
        md.add('bar', 1)
        md.add('bar', 2)
        self.assertEqual(md.to_dict(), {'foo': 42, 'bar': 1})
        self.assertEqual(md.to_dict(flat=False), {'foo': [42], 'bar': [1, 2]})

        # popitem from empty dict
        self.assertRaises(KeyError, self.storage_class().popitem)

        self.assertRaises(KeyError, self.storage_class().popitemlist)

        # key errors are of a special type
        self.assertRaises(BadRequestKeyError, lambda: self.storage_class()[42])

        # setlist works
        md = self.storage_class()
        md['foo'] = 42
        md.setlist('foo', [1, 2])
        self.assertEqual(md.getlist('foo'), [1, 2])


class _ImmutableDictTestsMixin(object):
    storage_class = None

    def test_follows_dict_interface(self):
        cls = self.storage_class

        data = {'foo': 1, 'bar': 2, 'baz': 3}
        d = cls(data)

        self.assertEqual(d['foo'], 1)
        self.assertEqual(d['bar'], 2)
        self.assertEqual(d['baz'], 3)
        self.assertEqual(sorted(d.keys()), ['bar', 'baz', 'foo'])
        self.assertIn('foo', d)
        self.assertNotIn('foox', d)
        self.assertEqual(len(d), 3)

    def test_copies_are_mutable(self):
        cls = self.storage_class
        immutable = cls({'a': 1})
        self.assertRaises(TypeError, immutable.pop, 'a')

        mutable = immutable.copy()
        mutable.pop('a')
        self.assertIn('a', immutable)
        self.assertIsNot(mutable, immutable)
        self.assertIs(copy(immutable), immutable)

    def test_dict_is_hashable(self):
        cls = self.storage_class
        immutable = cls({'a': 1, 'b': 2})
        immutable2 = cls({'a': 2, 'b': 2})
        x = set([immutable])
        self.assertIn(immutable, x)
        self.assertNotIn(immutable2, x)
        x.discard(immutable)
        self.assertNotIn(immutable, x)
        self.assertNotIn(immutable2, x)
        x.add(immutable2)
        self.assertNotIn(immutable, x)
        self.assertIn(immutable2, x)
        x.add(immutable)
        self.assertIn(immutable, x)
        self.assertIn(immutable2, x)


class ImmutableTypeConversionDictTestCase(
    _ImmutableDictTestsMixin, unittest.TestCase
):
    storage_class = datastructures.ImmutableTypeConversionDict


class ImmutableMultiDictTestCase(_ImmutableDictTestsMixin, unittest.TestCase):
    storage_class = datastructures.ImmutableMultiDict

    def test_multidict_is_hashable(self):
        cls = self.storage_class
        immutable = cls({'a': [1, 2], 'b': 2})
        immutable2 = cls({'a': [1], 'b': 2})
        x = set([immutable])
        self.assertIn(immutable, x)
        self.assertNotIn(immutable2, x)
        x.discard(immutable)
        self.assertNotIn(immutable, x)
        self.assertNotIn(immutable2, x)
        x.add(immutable2)
        self.assertNotIn(immutable, x)
        self.assertIn(immutable2, x)
        x.add(immutable)
        self.assertIn(immutable, x)
        self.assertIn(immutable2, x)


class ImmutableDictTestCase(
    _ImmutableDictTestsMixin, unittest.TestCase,
):
    storage_class = datastructures.ImmutableDict


class ImmutableOrderedMultiDictTestCase(
    _ImmutableDictTestsMixin, unittest.TestCase
):
    storage_class = datastructures.ImmutableOrderedMultiDict

    def test_ordered_multidict_is_hashable(self):
        a = self.storage_class([('a', 1), ('b', 1), ('a', 2)])
        b = self.storage_class([('a', 1), ('a', 2), ('b', 1)])
        self.assertNotEqual(hash(a), hash(b))


class MultiDictTestCase(_MutableMultiDictTestsMixin, unittest.TestCase):
    storage_class = datastructures.MultiDict

    def test_multidict_pop(self):
        def make_d():
            return self.storage_class({'foo': [1, 2, 3, 4]})
        d = make_d()
        self.assertEqual(d.pop('foo'), 1)
        self.assertFalse(d)
        d = make_d()
        self.assertEqual(d.pop('foo', 32), 1)
        self.assertFalse(d)
        d = make_d()
        self.assertEqual(d.pop('foos', 32), 32)
        self.assertTrue(d)

        self.assertRaises(KeyError, d.pop, 'foos')

    def test_setlistdefault(self):
        md = self.storage_class()
        self.assertEqual(md.setlistdefault('u', [-1, -2]), [-1, -2])
        self.assertEqual(md.getlist('u'), [-1, -2])
        self.assertEqual(md['u'], -1)

    def test_iter_interfaces(self):
        mapping = [
            ('a', 1), ('b', 2), ('a', 2), ('d', 3),
            ('a', 1), ('a', 3), ('d', 4), ('c', 3),
        ]
        md = self.storage_class(mapping)
        self.assertEqual(
            list(zip(md.keys(), md.listvalues())), list(md.lists())
        )
        self.assertEqual(
            list(zip(md, md.listvalues())), list(md.lists())
        )
        self.assertEqual(
            list(zip(md.keys(), md.listvalues())), list(md.lists())
        )


class OrderedMultiDictTestCase(_MutableMultiDictTestsMixin, unittest.TestCase):
    storage_class = datastructures.OrderedMultiDict

    def test_ordered_interface(self):
        cls = self.storage_class

        d = cls()
        self.assertFalse(d)
        d.add('foo', 'bar')
        self.assertEqual(len(d), 1)
        d.add('foo', 'baz')
        self.assertEqual(len(d), 1)
        self.assertEqual(list(d.items()), [('foo', 'bar')])
        self.assertEqual(list(d), ['foo'])
        self.assertEqual(
            list(d.items(multi=True)), [('foo', 'bar'), ('foo', 'baz')]
        )
        del d['foo']
        self.assertFalse(d)
        self.assertEqual(len(d), 0)
        self.assertEqual(list(d), [])

        d.update([('foo', 1), ('foo', 2), ('bar', 42)])
        d.add('foo', 3)
        self.assertEqual(d.getlist('foo'), [1, 2, 3])
        self.assertEqual(d.getlist('bar'), [42])
        self.assertEqual(list(d.items()), [('foo', 1), ('bar', 42)])

        expected = ['foo', 'bar']

        self.assertEqual(list(d.keys()), expected)
        self.assertEqual(list(d), expected)
        self.assertEqual(list(d.keys()), expected)

        self.assertEqual(list(d.items(multi=True)), [
            ('foo', 1), ('foo', 2), ('bar', 42), ('foo', 3),
        ])
        self.assertEqual(len(d), 2)

        self.assertEqual(d.pop('foo'), 1)
        self.assertIs(d.pop('blafasel', None), None)
        self.assertEqual(d.pop('blafasel', 42), 42)
        self.assertEqual(len(d), 1)
        self.assertEqual(d.poplist('bar'), [42])
        self.assertFalse(d)

        d.get('missingkey') is None

        d.add('foo', 42)
        d.add('foo', 23)
        d.add('bar', 2)
        d.add('foo', 42)
        self.assertEqual(d, datastructures.MultiDict(d))
        id = self.storage_class(d)
        self.assertEqual(d, id)
        d.add('foo', 2)
        self.assertNotEqual(d, id)

        d.update({'blah': [1, 2, 3]})
        self.assertEqual(d['blah'], 1)
        self.assertEqual(d.getlist('blah'), [1, 2, 3])

        # setlist works
        d = self.storage_class()
        d['foo'] = 42
        d.setlist('foo', [1, 2])
        self.assertEqual(d.getlist('foo'), [1, 2])

        self.assertRaises(BadRequestKeyError, d.pop, 'missing')
        self.assertRaises(BadRequestKeyError, lambda: d['missing'])

        # popping
        d = self.storage_class()
        d.add('foo', 23)
        d.add('foo', 42)
        d.add('foo', 1)
        self.assertEqual(d.popitem(), ('foo', 23))
        self.assertRaises(BadRequestKeyError, d.popitem)
        self.assertFalse(d)

        d.add('foo', 23)
        d.add('foo', 42)
        d.add('foo', 1)
        self.assertEqual(d.popitemlist(), ('foo', [23, 42, 1]))

        self.assertRaises(BadRequestKeyError, d.popitemlist)

    def test_iterables(self):
        a = datastructures.MultiDict((("key_a", "value_a"),))
        b = datastructures.MultiDict((("key_b", "value_b"),))
        ab = datastructures.CombinedMultiDict((a, b))

        self.assertEqual(sorted(ab.lists()), [
            ('key_a', ['value_a']), ('key_b', ['value_b']),
        ])
        self.assertEqual(sorted(ab.listvalues()), [['value_a'], ['value_b']])
        self.assertEqual(sorted(ab.keys()), ["key_a", "key_b"])


class CombinedMultiDictTestCase(unittest.TestCase):
    storage_class = datastructures.CombinedMultiDict

    def test_basic_interface(self):
        d1 = datastructures.MultiDict([('foo', '1')])
        d2 = datastructures.MultiDict([('bar', '2'), ('bar', '3')])
        d = self.storage_class([d1, d2])

        # lookup
        self.assertEqual(d['foo'], '1')
        self.assertEqual(d['bar'], '2')
        self.assertEqual(d.getlist('bar'), ['2', '3'])

        self.assertEqual(sorted(d.items()), [('bar', '2'), ('foo', '1')])
        self.assertEqual(
            sorted(d.items(multi=True)),
            [('bar', '2'), ('bar', '3'), ('foo', '1')]
        )
        self.assertNotIn('missingkey', d)
        self.assertIn('foo', d)

        # type lookup
        self.assertEqual(d.get('foo', type=int), 1)
        self.assertEqual(d.getlist('bar', type=int), [2, 3])

        # get key errors for missing stuff
        self.assertRaises(KeyError, lambda: d['missing'])

        # make sure that they are immutable
        try:
            d['foo'] = 'blub'
        except TypeError:
            pass
        else:
            self.fail()

        # copies are immutable
        d = d.copy()
        try:
            d['foo'] = 'blub'
        except TypeError:
            pass
        else:
            self.fail()

        # make sure lists merges
        md1 = datastructures.MultiDict((("foo", "bar"),))
        md2 = datastructures.MultiDict((("foo", "blafasel"),))
        x = self.storage_class((md1, md2))
        self.assertEqual(list(x.lists()), [('foo', ['bar', 'blafasel'])])

    def test_length(self):
        d1 = datastructures.MultiDict([('foo', '1')])
        d2 = datastructures.MultiDict([('bar', '2')])
        self.assertEqual(len(d1) == len(d2), 1)
        d = self.storage_class([d1, d2])
        self.assertEqual(len(d), 2)
        d1.clear()
        self.assertEqual(len(d1), 0)
        self.assertEqual(len(d), 1)


class ImmutableListTestCase(unittest.TestCase):
    storage_class = datastructures.ImmutableList

    def test_list_hashable(self):
        t = (1, 2, 3, 4)
        l = self.storage_class(t)
        self.assertEqual(hash(t), hash(l))
        self.assertNotEqual(t, l)


def make_call_asserter(func=None):
    """Utility to assert a certain number of function calls.

    :param func:
        Additional callback for each function call.

    >>> assert_calls, func = make_call_asserter()
    >>> with assert_calls(2):
            func()
            func()
    """

    calls = 0

    @contextmanager
    def asserter(count, msg=None):
        nonlocal calls
        calls = 0
        yield
        assert calls == count

    def wrapped(*args, **kwargs):
        nonlocal calls
        calls += 1
        if func is not None:
            return func(*args, **kwargs)

    return asserter, wrapped


class CallbackDictTestCase(unittest.TestCase):
    storage_class = datastructures.CallbackDict

    def test_callback_dict_reads(self):
        assert_calls, func = make_call_asserter()
        initial = {'a': 'foo', 'b': 'bar'}
        dct = self.storage_class(initial=initial, on_update=func)
        with assert_calls(0, 'callback triggered by read-only method'):
            # read-only methods
            dct['a']
            dct.get('a')
            self.assertRaises(KeyError, lambda: dct['x'])
            'a' in dct
            list(iter(dct))
            dct.copy()
        with assert_calls(0, 'callback triggered without modification'):
            # methods that may write but don't
            dct.pop('z', None)
            dct.setdefault('a')

    def test_callback_dict_writes(self):
        assert_calls, func = make_call_asserter()
        initial = {'a': 'foo', 'b': 'bar'}
        dct = self.storage_class(initial=initial, on_update=func)
        with assert_calls(8, 'callback not triggered by write method'):
            # always-write methods
            dct['z'] = 123
            dct['z'] = 123  # must trigger again
            del dct['z']
            dct.pop('b', None)
            dct.setdefault('x')
            dct.popitem()
            dct.update([])
            dct.clear()
        with assert_calls(0, 'callback triggered by failed del'):
            self.assertRaises(KeyError, lambda: dct.__delitem__('x'))
        with assert_calls(0, 'callback triggered by failed pop'):
            self.assertRaises(KeyError, lambda: dct.pop('x'))



"""
    verktyg.testsuite.dispatch
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    Tests for the core dispatcher.

    :copyright: (c) 2014 by Ben Mather.
    :license: BSD, see LICENSE for more details.
"""
import unittest

from verktyg.exceptions import NotImplemented, MethodNotAllowed, NotAcceptable
from verktyg.dispatch import Binding, Dispatcher


def _make_view(value):
    def view(app, req):
        return value
    return view


class DispatchTestCase(unittest.TestCase):
    def test_name_dispatch(self):
        dispatcher = Dispatcher([
            Binding('tweedle-dum', _make_view('Tweedle Dum')),
            Binding('tweedle-dee', _make_view('Tweedle Dee')),
            Binding('same', _make_view('overridden')),
            Binding('same', _make_view('overriding')),
        ])

        self.assertEqual(
            'Tweedle Dum', dispatcher.lookup('tweedle-dum')(None, None)
        )
        self.assertEqual(
            'Tweedle Dee', dispatcher.lookup('tweedle-dee')(None, None)
        )
        self.assertRaises(NotImplemented, dispatcher.lookup, 'non-existant')
        self.assertEqual(
            'overriding', dispatcher.lookup('same')(None, None)
        )

    def test_method_dispatch(self):
        dispatcher = Dispatcher([
            Binding('test', _make_view('get'), method='GET'),
            Binding('test', _make_view('post'), method='POST'),
            Binding('head', _make_view('head'), method='HEAD'),
            Binding('no-head', _make_view('get'), method='GET'),
        ])

        # default to 'GET'
        self.assertEqual(
            'get', dispatcher.lookup('test')(None, None)
        )
        self.assertEqual(
            'get', dispatcher.lookup('test', method='GET')(None, None)
        )

        # `POST` gives something different
        self.assertEqual(
            'post', dispatcher.lookup('test', method='POST')(None, None)
        )

        # `PUT` not found
        self.assertRaises(
            MethodNotAllowed, dispatcher.lookup, 'test', method='PUT'
        )

    def test_head_fallback(self):
        dispatcher = Dispatcher([
            Binding('head', _make_view('head'), method='HEAD'),
            Binding('no-head', _make_view('get'), method='GET'),
        ])

        # `HEAD` should fall back to `GET`
        self.assertEqual(
            'head', dispatcher.lookup('head', method='HEAD')(None, None)
        )
        self.assertEqual(
            'get', dispatcher.lookup('no-head', method='HEAD')(None, None)
        )

    def test_method_override(self):
        dispatcher = Dispatcher([
            Binding('same', _make_view('overridden')),
            Binding('same', _make_view('unaffected'), method='POST'),
            Binding('same', _make_view('overriding')),
        ])

        # replacing handler for one method should not affect others
        self.assertEqual(
            'overriding', dispatcher.lookup('same')(None, None)
        )
        self.assertEqual(
            'unaffected', dispatcher.lookup('same', method='POST')(None, None)
        )

    def test_accept_dispatch(self):
        dispatcher = Dispatcher([
            Binding(
                'test', _make_view('json'),
                content_type='application/json'
            ),
            Binding(
                'test', _make_view('html'),
                content_type='text/html'
            ),
            Binding(
                'test', _make_view('whatever')
            ),
            Binding(
                'nope', _make_view('nope'),
                content_type='application/xml'
            ),
        ])

        # accept header strings
        self.assertEqual(
            'json',
            dispatcher.lookup('test', accept='application/json')(None, None)
        )

        self.assertEqual(
            'json',
            dispatcher.lookup(
                'test', accept='application/json; q=0.9, text/html; q=0.8'
            )(None, None)
        )
        self.assertEqual(
            'whatever',
            dispatcher.lookup('test', accept='application/xml')(None, None)
        )

        self.assertRaises(
            NotAcceptable,
            dispatcher.lookup, 'nope', accept='text/html'
        )

    def test_nested(self):
        child = Dispatcher([
            Binding('nested', _make_view('Nested')),
        ])
        parent = Dispatcher([
            child,
        ])

        self.assertEqual(
            'Nested',
            parent.lookup('nested')(None, None))
